// Fvfo: 4700 - 5100 KHz
// Frf:  7300 - 6900 KHz 
//
// used Arduino board: e.g. "ProMini" 3V or similar
//  
// used Si5351A breakout board ( Fx=27MHz ): 
// http://www.qrp-labs.com/synth.html
//
// original software code made by OE1CGS, extended by DJ7OO
// http://www.qrp-labs.com/synth/oe1cgs.html 
//
// OLED: 4pin I2C 0.96"
// 
// OLED and Si5351A I2C SDA: #A4
//                      SCL: #A5 
//
// STEP selection     by connecting #8 to ground 
// EEPROM FRQ storing by connecting #9 to ground 

#include <Wire.h>
#include "U8glib.h"
#include "EEPROM.h" 

U8GLIB_SSD1306_128X64 u8g(U8G_I2C_OPT_NONE);

 float frq; 
 float frq_old;
 float step = 1000;
 float step_old; 
 unsigned long currentTime;
 unsigned long loopTime;
 const int pin_A = 2;  // pin 2
 const int pin_B = 3;  // pin 3
 unsigned char encoder_A;
 unsigned char encoder_B;
 unsigned char encoder_A_prev=0;
 int press = 0;
 int cnt = 1;
 boolean toggle = true;

/////////////////////////////////////////////////////////////////////////
void Si5351a_Write_Reg (byte regist, byte value)
{                                                   // Writes "byte" into "regist" of Si5351a via I2C
  Wire.beginTransmission(96);                       // Starts transmission as master to slave 96, which is the
                                                    // I2C address of the Si5351a (see Si5351a datasheet)
  Wire.write(regist);                               // Writes a byte containing the number of the register
  Wire.write(value);                                // Writes a byte containing the value to be written in the register
  Wire.endTransmission();                           // Sends the data and ends the transmission
}

/////////////////////////////////////////////////////////////
void SetFrequency (unsigned long frequency) { // Frequency in Hz; must be within [7810 Hz to 200 Mhz]
  #define F_XTAL 27005701;             // Frequency of Quartz-Oszillator
  #define c 1048574;                   // "c" part of Feedback-Multiplier from XTAL to PLL
  unsigned long fvco;                  // VCO frequency (600-900 MHz) of PLL
  unsigned long outdivider;            // Output divider in range [4,6,8-900], even numbers preferred
  byte R = 1;                          // Additional Output Divider in range [1,2,4,...128]
  byte a;                              // "a" part of Feedback-Multiplier from XTAL to PLL in range [15,90]
  unsigned long b;                     // "b" part of Feedback-Multiplier from XTAL to PLL
  float f;                             // floating variable, needed in calculation
  unsigned long MS0_P1;                // Si5351a Output Divider register MS0_P1, P2 and P3 are hardcoded below
  unsigned long MSNA_P1;               // Si5351a Feedback Multisynth register MSNA_P1
  unsigned long MSNA_P2;               // Si5351a Feedback Multisynth register MSNA_P2
  unsigned long MSNA_P3;               // Si5351a Feedback Multisynth register MSNA_P3
  outdivider = 900000000 / frequency;  // With 900 MHz beeing the maximum internal PLL-Frequency  
  while (outdivider > 900){            // If output divider out of range (>900) use additional Output divider
    R = R * 2;
    outdivider = outdivider / 2;
  }
  if (outdivider % 2) outdivider--;    // finds the even divider which delivers the intended Frequency
  fvco = outdivider * R * frequency;   // Calculate the PLL-Frequency (given the even divider)
  switch (R){                          // Convert the Output Divider to the bit-setting required in register 44
    case 1: R = 0; break;              // Bits [6:4] = 000
    case 2: R = 16; break;             // Bits [6:4] = 001
    case 4: R = 32; break;             // Bits [6:4] = 010
    case 8: R = 48; break;             // Bits [6:4] = 011
    case 16: R = 64; break;            // Bits [6:4] = 100
    case 32: R = 80; break;            // Bits [6:4] = 101
    case 64: R = 96; break;            // Bits [6:4] = 110
    case 128: R = 112; break;          // Bits [6:4] = 111
  }
  a = fvco / F_XTAL;                   // Multiplier to get from Quartz-Oscillator Freq. to PLL-Freq.
  f = fvco - a * F_XTAL;               // Multiplier = a+b/c
  f = f * c;                           // this is just "int" and "float" mathematics
  f = f / F_XTAL;
  b = f;
  MS0_P1 = 128 * outdivider - 512;     // Calculation of Output Divider registers MS0_P1 to MS0_P3
                                       // MS0_P2 = 0 and MS0_P3 = 1; these values are hardcoded, see below
  f = 128 * b / c;                     // Calculation of Feedback Multisynth registers MSNA_P1 to MSNA_P3
  MSNA_P1 = 128 * a + f - 512;
  MSNA_P2 = f;
  MSNA_P2 = 128 * b - MSNA_P2 * c; 
  MSNA_P3 = c;
  Si5351a_Write_Reg (16, 128);                      // Disable output during the following register settings 
  Si5351a_Write_Reg (26, (MSNA_P3 & 65280) >> 8);   // Bits [15:8] of MSNA_P3 in register 26
  Si5351a_Write_Reg (27, MSNA_P3 & 255);            // Bits [7:0]  of MSNA_P3 in register 27
  Si5351a_Write_Reg (28, (MSNA_P1 & 196608) >> 16); // Bits [17:16] of MSNA_P1 in bits [1:0] of register 28
  Si5351a_Write_Reg (29, (MSNA_P1 & 65280) >> 8);   // Bits [15:8]  of MSNA_P1 in register 29
  Si5351a_Write_Reg (30, MSNA_P1 & 255);            // Bits [7:0]  of MSNA_P1 in register 30
  Si5351a_Write_Reg (31, ((MSNA_P3 & 983040) >> 12) | ((MSNA_P2 & 983040) >> 16)); // Parts of MSNA_P3 und MSNA_P1
  Si5351a_Write_Reg (32, (MSNA_P2 & 65280) >> 8);   // Bits [15:8]  of MSNA_P2 in register 32
  Si5351a_Write_Reg (33, MSNA_P2 & 255);            // Bits [7:0]  of MSNA_P2 in register 33
  Si5351a_Write_Reg (42, 0);                        // Bits [15:8] of MS0_P3 (always 0) in register 42
  Si5351a_Write_Reg (43, 1);                        // Bits [7:0]  of MS0_P3 (always 1) in register 43
  Si5351a_Write_Reg (44, ((MS0_P1 & 196608) >> 16) | R);  // Bits [17:16] of MS0_P1 in bits [1:0] and R in [7:4]
  Si5351a_Write_Reg (45, (MS0_P1 & 65280) >> 8);    // Bits [15:8]  of MS0_P1 in register 45
  Si5351a_Write_Reg (46, MS0_P1 & 255);             // Bits [7:0]  of MS0_P1 in register 46
  Si5351a_Write_Reg (47, 0);                        // Bits [19:16] of MS0_P2 and MS0_P3 are always 0
  Si5351a_Write_Reg (48, 0);                        // Bits [15:8]  of MS0_P2 are always 0
  Si5351a_Write_Reg (49, 0);                        // Bits [7:0]   of MS0_P2 are always 0
  if (outdivider == 4){
    Si5351a_Write_Reg (44, 12 | R);                 // Special settings for R = 4 (see datasheet)
    Si5351a_Write_Reg (45, 0);                      // Bits [15:8]  of MS0_P1 must be 0
    Si5351a_Write_Reg (46, 0);                      // Bits [7:0]  of MS0_P1 must be 0
  } 
  Si5351a_Write_Reg (177, 32);                      // This resets PLL A
}

///////////////////////////////////////////////////////////////////
void rotary_enc()
{
   currentTime = millis();
   if(currentTime >= (loopTime + 2)){
    encoder_A = digitalRead(pin_A);    
    encoder_B = digitalRead(pin_B);   
    if((!encoder_A) && (encoder_A_prev)){
    if(encoder_B) {
        frq = frq+step;               
        if(frq>5100000) {frq=4700000;}   
      }   
      else {
        frq = frq-step;               
        if(frq<4700000) {frq=5100000;} 
      }   
    }   
    encoder_A_prev = encoder_A;     // Store value of A for next time       
    loopTime = currentTime;         // Updates loopTime
    }
  }

//////////////////////////////////////////////////////////////
void EEPROMWritelong(int address,long value)
      {
      //Decomposition from a long to 4 bytes by using bitshift.
      //One = Most significant -> Four = Least significant byte
      byte four = (value & 0xFF);
      byte three = ((value >> 8) & 0xFF);
      byte two = ((value >> 16) & 0xFF);
      byte one = ((value >> 24) & 0xFF);

      //Write the 4 bytes into the eeprom memory.
      EEPROM.write(address, four);
      EEPROM.write(address + 1, three);
      EEPROM.write(address + 2, two);
      EEPROM.write(address + 3, one);
      }

//////////////////////////////////////////////////////////////
//This function will return a 4 byte (32bit) long from the eeprom
//at the specified address to address + 3.
long EEPROMReadlong(long address)
      {
      //Read the 4 bytes from the eeprom memory.
      long four = EEPROM.read(address);
      long three = EEPROM.read(address + 1);
      long two = EEPROM.read(address + 2);
      long one = EEPROM.read(address + 3);

      //Return the recomposed long by using bitshift.
      return ((four << 0) & 0xFF) + ((three << 8) & 0xFFFF) + ((two << 16) & 0xFFFFFF) + ((one << 24) & 0xFFFFFFFF);
      }

////////////////////////////////////////////////////////////  
void oled()
  {
  u8g.firstPage();  
  do 
  {
  u8g.drawFrame(1,18,127,25);
  u8g.setFont(u8g_font_ncenB10);
  u8g.setPrintPos(1,12);
  u8g.print("RF-FREQUENCY"); 

  u8g.setPrintPos(5,62); 
  u8g.print("Step: ");
  if(step<1000) 
{ u8g.print("100 Hz"); }
  else  
{ u8g.print(step/1000,0); 
  u8g.print(" KHz"); }  

{ u8g.setPrintPos(7,40); 
  u8g.setFont(u8g_font_ncenB18);       
  u8g.print(12000-(frq/1000),1);
  u8g.setFont(u8g_font_ncenB10);
  u8g.print(" KHz"); }
  }
  while ( u8g.nextPage() );  
  }

///////////////////////////////////////////////////////////////
void setup() 
 {
 Serial.begin(9600);   
 Serial.println("Start");
 delay(100);
 pinMode(8,INPUT_PULLUP);  
 pinMode(9,INPUT_PULLUP);  

  u8g.firstPage();  
   do {
  u8g.drawFrame(1,30,127,32);
  u8g.setFont(u8g_font_ncenB10);
  u8g.setPrintPos(15,15);
  u8g.print(" BITX40-VFO "); 
  u8g.setFont(u8g_font_ncenB10);
  u8g.setPrintPos(10,50); 
  u8g.print("(C)DJ7OO 2016");
  } 
  while ( u8g.nextPage() );

 pinMode(pin_A, INPUT_PULLUP);
 pinMode(pin_B, INPUT_PULLUP);
 currentTime = millis();
 loopTime = currentTime; 
 frq = EEPROMReadlong(0); 
 Serial.print("Frf read from EEPROM; ");
 Serial.print(frq/1000,1);
 Serial.println(" KHz"); 
 if(frq<4700000||frq>5100000) {frq=4850000;} 

 Wire.begin();                         // Initialize I2C-communication as master
                                       // SDA on pin ADC04
                                       // SCL on pin ADC05
  SetFrequency(frq);                   // Set TX-Frequency 
  Si5351a_Write_Reg (17, 128);         // Disable output CLK1
  Si5351a_Write_Reg (16, 79);          // Enable output CLK0, set crystal as source and Integer Mode on PLLA 
  delay(3000);
}

///////////////////////////////////////////////////////////////
void loop() 
{
  press = digitalRead(9);
  if (press == LOW)
  {
  cnt = cnt+1; 
  if(cnt==3) { cnt=0 ; }
//  Serial.println(cnt);
  if(cnt==0) { step=100; }
  else
  if(cnt==1)  { step=1000; }
  else
  if(cnt==2) { step=10000; }
  delay(300);
  }

  press = digitalRead(8);
  if (press == LOW)
  {
  long address=0;
  EEPROMWritelong(address, frq);
  address+=4;
  Serial.print("Frf stored in EEPROM: ");
  Serial.print(frq/1000,1);
  Serial.println(" KHz");
  delay(300);
  }
  
  rotary_enc();
  if(frq!=frq_old)
  {
  SetFrequency(frq);                   // Set Frequency 
  Serial.print("Fosz: ");
  Serial.print(frq/1000,1);
  Serial.print(" KHz ");
  Serial.print("   Frf:  ");
  Serial.print(12000-(frq/1000),1);
  Serial.println(" KHz ");
  Si5351a_Write_Reg (17, 128);         // Disable output CLK1
  Si5351a_Write_Reg (16, 79);          // Enable output CLK0, set crystal as source and Integer Mode on PLLA
                                       // CLK0 drive strength = 8mA; power level := 0dB
  oled();
  }
  if(step!=step_old) { oled(); }   
  frq_old = frq;
  step_old = step;
 }
 
//////////////////////////////////////////////////////////

